'' =================================================================================================
''
''   File....... jm_wav_player_2015.spin
''   Purpose.... Mono/Stereo player for 16-bit canonical WAV files
''               -- file must be free of metadata
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2009-2015 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....
''   Started....
''   Updated.... 12 FEB 2015
''
'' =================================================================================================


con { fixed io pins }

  RX  = 31                                                       ' programming port
  TX  = 30

  SDA = 29                                                       ' i2c port
  SCL = 28


con

  BUF_SIZE = 512                                                 ' wav data buffer size

  VOL_MAX  = 100
  VOL_MIN  =   0


con { error codes }

  NO_FILE    =  -1                                               ' file not found on sd

  BAD_FILE   =  -2                                               ' file not compatible with player
  BAD_HDR    =  -3                                               ' error in WAV header
  BAD_FORMAT =  -4                                               ' incorrect WAV format
  BAD_CHANS  =  -5                                               ' bad channel count
  BAD_RATE   =  -6                                               ' incompatible sample rate
  BAD_BITS   =  -7                                               ' bad sample size

  NO_MOUNT   = -10                                               ' sd card failed to mount

  NO_COG     = -20                                               ' no cog(s) available for player


obj

  dacs : "stereo_dacs"                                           ' load into cog 1 if possible
  sd   : "fsrw"


var

  long  daccog                                                   ' stereo_duy    (PASM)
  long  sdcog                                                    ' sd spi cog    (PASM)
  long  wavcog                                                   ' wav player    (PASM)
  long  bufcog                                                   ' double buffer (SPIN)
  long  bufstack[128]                                            ' stack for spin cog

  long  lpin                                                     ' do not change order of this group
  long  rpin                                                     '
  long  sddo                                                     '
  long  sdclk                                                    '
  long  sddi                                                     '
  long  sdcs                                                     '
  long  dither                                                   '

  long  lrsamples                                                ' for stereo_duty

  long  mounted                                                  ' true when SD ready

  long  playflag                                                 ' playing
  long  filerate                                                 ' original sample rate
  long  srate                                                    ' sample rate (ticks)
  long  buf0count                                                ' # samples in buf0
  long  buf0[BUF_SIZE]                                           ' samples buffer (from SD)
  long  buf1count
  long  buf1[BUF_SIZE]

  long  lvolume                                                  ' left volume (0 to 100%)
  long  rvolume                                                  ' right volume (0 to 100%)
  long  speed                                                    ' playback, 50 to 100%

  byte  header[44]                                               ' for WAV header

  byte  filename[13]                                             ' for file name


pub start(p_pindefs) | check

'' Start WAV playback cogs
'' -- p_pindefs is pointer (array of longs)

  stop                                                           ' stop if running

  longmove(@lpin, p_pindefs, 7)                                  ' get pins

  daccog := dacs.start(@lrsamples, lpin, rpin, dither)           ' start dac driver
  if (daccog == false)
    return NO_COG

  ' set wav cog parameters
  ' -- update PASM code before moving to cog

  sdpntr    := @lrsamples
  pfpntr    := @playflag
  srpntr    := @srate
  b0cntpntr := @buf0count
  b1cntpntr := @buf1count
  b0pntr    := @buf0
  b1pntr    := @buf1
  lvolpntr  := @lvolume
  rvolpntr  := @rvolume

  set_volume(MAX_VOL, MAX_VOL)                                   ' default to max
  set_speed(100)                                                 ' default norm

  check := mount_sd
  if (check == 0)
    wavcog := cognew(@wavdacs, 0) + 1                            ' start wav driver cog
    if (wavcog)                                                  ' did it load?
      return 0                                                   ' yes, return 0
    else
      stop                                                       ' no, unload stereo duty
      return NO_COG
  else
    return NO_MOUNT


pub stop

'' Stop cogs running WAV player elements

  kill                                                           ' kill audio if playing

  if (wavcog)                                                    ' wav driver
    cogstop(wavcog - 1)
    wavcog := 0

  if (sdcog)                                                     ' safe_spi for fsrw
    \sd.unmount
    sdcog := 0

  if (daccog)                                                    ' stereo dac cog
    cogstop(daccog - 1)
    daccog := 0


pri mount_sd | check

'' Attempts to mount SD card

  check := xmount(sddo, sdclk, sddi, sdcs)

  if (check <> 0)                                                ' if problem...
    sdcog := false                                               '  mark not mounted
    mounted := false
    \sd.unmount                                                  '  unload safe_spi
    return NO_MOUNT
  else
    sdcog := true                                                ' else mark good mount
    mounted := true
    return 0


pub sd_mounted

'' Returns true if SD card is mounted

  return mounted


pub set_volume(l, r)

'' Set playback levels for left and right channels, 0 to 100%

  l := (VOL_MIN #> l <# VOL_MAX)                                 ' keep in legal range
  if (l == 100)
    lvolume := $1_0000                                           ' avoid rounding errors
  else
    lvolume := l * $1_0000 / 100

  r := (VOL_MIN #> r <# VOL_MAX)
  if (r == 100)
    rvolume := $1_0000
  else
    rvolume := r * $1_0000 / 100


pub set_speed(spd)

'' Set playback speed, 50% to 200%
'' -- 100MHz clock (6.25MHz x 16) suggested for 48kHz files

  speed := 50 #> spd <# 200
  srate := ((clkfreq / filerate) * 100 / speed) #> (clkfreq / 48_000)


pub play(p_filename, n) | check

'' Starts file in spntr, n times
'' -- 0 to loop "forever"
'' -- will stop player if active, even if bad file specifed

  stop_player                                                    ' stop if playing now

  if (mounted == false)                                          ' is card mounted?
    check := mount_sd                                               ' no, attempt
    if (check <> true)                                           ' if mount error
      return NO_MOUNT                                            ' exit with error code

  check := has_file(p_filename)
  if (check <> 0)
    return NO_FILE

  check := check_file(p_filename, false)                         ' check file specs
  if (check == 0)                                                ' okay?
    bufcog := cognew(wav_spool(p_filename, n), @bufstack) + 1    ' yes, start file
    if (bufcog == 0)
      return NO_COG
    else
      return 0
  else
    return BAD_FILE


pub playing

'' Returns player status
'' -- When playging, bufcog (cog running wav data double-buffer) is non-zero

  return (bufcog > 0)


pub kill

'' Stops audio buffering cog if active

  if (bufcog)
    stop_player                                                  ' shutdown/clear buffers
    pclose                                                       ' close file
    cogstop(bufcog - 1)
    bufcog := 0


pub has_file(p_filename) | check

'' Searches for file (at p_zstr) on SD card
''   0 -- file present
''  -1 -- not found

  stop_player                                                    ' stop if playing

  check := \sd.popen(p_filename, "r")                            ' try to open
  if (check <> 0)
    return NO_FILE                                               ' file not found

  return 0


pub get_header(p_filename) | check

'' Opens file at p_filename, moves header (44 bytes) into local buffer

  bytefill(@header, 0, 44)                                       ' empty buffer

  check := popen(p_filename, "r")                                ' attempt to open
  if (check <> 0)
    pclose
    return NO_FILE

  check := pread(@header, 44)                                    ' get header bytes
  if (check <> 44)
    pclose
    return BAD_HDR

  return 0


pub format | fmt

'' Returns format of WAV file

  fmt := 0
  wordmove(@fmt, @header[20], 1)

  return fmt


pub channels | ch

'' Returns channels in WAV file

  ch := 0
  wordmove(@ch, @header[22], 1)

  return ch


pub sample_rate

'' Returns sample rate (Hz) of WAV file
'' -- uses last opened file

  longmove(@filerate, @header[24], 1)

  return filerate


pub bits_per_sample | bits

'' Returns bits per sample of WAV file

  bits := 0
  wordmove(@bits, @header[34], 1)

  return bits


pub samples | balign, chunk2

'' Returns samples in WAV file (chunk2)

  balign := 0
  wordmove(@balign, @header[32], 1)                              ' block align (bytes per sample)
  chunk2 := 0
  longmove(@chunk2, @header[40], 1)                              ' bytes in chunk2 (audio)

  return chunk2 / balign


pub audio_bytes | chunk2

  longmove(@chunk2, @header[40], 1)                              ' bytes in chunk2 (audio)

  return chunk2


pub duration

'' Returns duration of audio chunk in 1/100ths seconds

  return (samples * 10) / (sample_rate / 10)


pub check_file(p_filename, remount) | check

'' Ensures file is present and correct type (PCM, stereo, 16-bit)
'' -- must stop audio in progress to check file

  stop_player                                                    ' kill if running now

  if ((remount == true) or (mounted == false))
    check := mount_sd
    if (check < 0)
      return NO_MOUNT

  check := popen(p_filename, "r")                                ' try to open
  if (check <> 0)
    sd.pclose
    return NO_FILE

  check := pread(@header, 44)
  sd.pclose
  if (check <> 44)
    return BAD_HDR                                               ' bad file

  if (format <> 1)
    return BAD_FORMAT                                            ' not pcm

  if ((channels <> 1) and (channels <> 2))
    return BAD_CHANS                                             ' not mono or stereo

  if ((sample_rate < 2_000) or (sample_rate > 48_000))
    return BAD_RATE                                              ' bad sample rate

  if (bits_per_sample <> 16)
    return BAD_BITS                                              ' incorrect bit size

  return 0                                                       ' no problems


pri stop_player

'' Stops player
'' -- clears play flag, rate, and play buffers

  playflag := false                                              ' stop playing
  srate := 0                                                     ' clear rate
  clear_bufs                                                     ' clear buffers


pri start_player(rate) | tmp

'' Sets sample rate for driver and enables play flag

  if (speed == 100)
    srate := clkfreq / rate
  else
    srate := ((clkfreq / rate) * 100 / speed) #> (clkfreq / 48_000)

  playflag := channels                                          ' play now


pri wav_spool(p_filename, plays) | n, playbytes, playcount

'' Spools WAV data to DACs engine
'' -- this method is launched into its own cog

  if (plays =< 0)
    plays := posx                                                ' if 0, play "forever"

  repeat plays
    popen(p_filename, "r")                                       ' open the file
    pread(@header, 44)                                           ' read the header
    start_player(sample_rate)                                    ' enable the player

    playbytes := audio_bytes                                     ' play only audio data
    playcount := 0

    repeat
      if (buf0count == 0)                                        ' if buf0 is empty
        n := sd.pread(@buf0, constant(BUF_SIZE << 2))            ' read from file
        if (n > 0)                                               ' have bytes?
          buf0count := n >> 2                                    ' yes, convert to samples
          playcount += n
          if (playcount => playbytes)
            quit
        else                                                     ' at end of file
          buf0count := 0
          buf1count := 0
          quit                                                   ' we're done

      if (buf1count == 0)
        n := sd.pread(@buf1, constant(BUF_SIZE << 2))
        if (n > 0)
          buf1count := n >> 2
          playcount += n
          if (playcount => playbytes)
            quit
        else
          buf0count := 0
          buf1count := 0
          quit

    repeat until ((buf0count == 0) and (buf1count == 0))        ' let buffers run out

    sd.pclose                                                   ' close file
    stop_player                                                 ' shutdown/clear buffers

  stop_player
  bufcog := 0                                                   ' alert foreground
  cogstop(cogid)                                                ' unload this cog


pri clear_bufs

  longfill(@buf0count, 0, (1 + BUF_SIZE) << 1)
  bytefill(@header, 0, 44)
  bytefill(@filename, 0, 13)


con

  ' shell methods for fsrw features
  ' -- assumes mounted SD card
  ' -- not all features implemented


pub  xmount(do, clk, di, cs)

  return \sd.mount_explicit(do, clk, di, cs)


pub popen(spntr, cmd)

  return \sd.popen(spntr, cmd)


pub pclose

  return \sd.pclose


pub get_filesize

  return \sd.get_filesize


pub pread(ubuf, count)

  return \sd.pread(ubuf, count)


pub pgetc

  return \sd.pgetc


pub pwrite(ubuf, count)

  return \sd.pwrite(ubuf, count)


pub pputs(p_zstr)

  return \sd.pputs(p_zstr)


dat { wav driver }

                        org     0

wavdacs
quiet                   wrlong  ZERO, sdpntr                     ' write silence to Stereo_Duty

                        rdlong  mode, pfpntr            wz       ' read play flag
        if_z            jmp     #quiet

                        rdlong  ratetix, srpntr         wz       ' sample rate tix
        if_z            jmp     #quiet

                        call    #check0                          ' anything in 1st buffer?
        if_z            jmp     #quiet

                        ' ready to play

                        mov     smpltimer, cnt
                        add     smpltimer, ratetix

playloop                rdlong  mode, pfpntr            wz       ' check play flag
        if_z            jmp     #quiet

                        test    mode, #%10              wc       ' test channels
        if_c            jmp     #playstereo


' **********************
' M O N O
' **********************
'
playmono                waitcnt smpltimer, ratetix               ' let sample timer expire

                        rdlong  msamples, bufpntr                ' get two samples
                        mov     lchan, msamples                  ' copy to left
                        mov     rchan, msamples                  ' copy to right

                        rdlong  lvol, lvolpntr                   ' get volume settings

                        and     lchan, hx0000_FFFF               ' low sample 1st
                        and     rchan, hx0000_FFFF               ' low sample 1st

'                       rdlong  rvol, rvolpntr                   ' not used

                        rdlong  ratetix, srpntr         wz       ' sample rate tix
        if_z            jmp     #quiet

sample1l                mov     arg1, lchan
                        mov     arg2, lvol
                        call    #adjust                          ' adjust for volume
                        mov     lchan, arg1

sample1r                mov     arg1, rchan
                        mov     arg2, rvol
                        call    #adjust
                        mov     rchan, arg1

                        shl     rchan, #16                       ' move to high word
                        or      lchan, rchan                     ' reconstruct
                        wrlong  lchan, sdpntr                    ' update Stereo_Duty

                        waitcnt smpltimer, ratetix               ' let sample timer expire

sample2l                mov     lchan, msamples                  ' copy to left
                        shr     lchan, #16                       ' high sample 2nd
                        mov     arg1, lchan
                        mov     arg2, lvol
                        call    #adjust                          ' adjust for volume
                        mov     lchan, arg1

sample2r                mov     rchan, msamples                  ' copy to right
                        shr     rchan, #16
                        mov     arg1, rchan
                        mov     arg2, rvol
                        call    #adjust
                        mov     rchan, arg1

                        shl     rchan, #16                       ' move to high word
                        or      lchan, rchan                     ' reconstruct
                        wrlong  lchan, sdpntr                    ' update Stereo_Duty

                        jmp     #updatebuffer


' **********************
' S T E R E O
' **********************
'
playstereo              waitcnt smpltimer, ratetix               ' let sample timer expire

                        rdlong  lchan, bufpntr                   ' get left & right
                        mov     rchan, lchan                     ' copy right

                        rdlong  lvol, lvolpntr                   ' get volume settings
                        rdlong  rvol, rvolpntr

cleanup                 and     lchan, hx0000_FFFF               ' restore to 16-bit
                        shr     rchan, #16

                        rdlong  ratetix, srpntr         wz       ' sample rate tix
        if_z            jmp     #quiet

adjleft                 mov     arg1, lchan
                        mov     arg2, lvol
                        call    #adjust
                        mov     lchan, arg1

adjright                mov     arg1, rchan
                        mov     arg2, rvol
                        call    #adjust
                        mov     rchan, arg1

                        shl     rchan, #16                       ' move to high word
                        or      lchan, rchan                     ' reconstruct
                        wrlong  lchan, sdpntr                    ' update Stereo_Duty


updatebuffer            add     bufpntr, #4                      ' point to next sample
                        djnz    bcount, #playloop                ' done with this buffer?

                        wrlong  ZERO, bcpntr                     ' clear this buffer for re-fill
                        test    bufid, #1               wc       ' which buffer just played?

        if_c            call    #check0                          ' anything in buffer 0?
        if_nc           call    #check1                          ' anything in buffer 1?
        if_nz           jmp     #playloop                        ' we're still playing

                        wrlong  ZERO, pfpntr                     ' clear play flag
                        jmp     #quiet


' check samples in buffer
' -- resets buffer id and pointer to buffer

check0                  mov     bufid, #0                        ' set buffer id
                        mov     bcpntr, b0cntpntr                ' point to count
                        rdlong  bcount, bcpntr          wz       ' get buffer samples count
                        mov     bufpntr, b0pntr                  ' point to element 0
check0_ret              ret


check1                  mov     bufid, #1
                        mov     bcpntr, b1cntpntr
                        rdlong  bcount, bcpntr          wz
                        mov     bufpntr, b1pntr
check1_ret              ret



' Multiply signed, 16-bit channel data by unsigned (16-bit) adjustment
' -- adjustment expressed as percent (0.00 to 1.00) x $1_0000
'
' arg1 = channel value in (signed, 16-bit)
' arg2 = adjustment (unsigned)
' t1 = holds sign bit
' t2 = working variable

'
adjust                  cmp     arg2, MAX_VOL           wc, wz   ' 100%?
        if_ae           jmp     #adjust_ret

                        cmps    arg2, #1                wc       ' 0%
        if_b            mov     arg1, #0
        if_b            jmp     #adjust_ret

                        shl     arg1, #16                        ' extend sign from bit15
                        sar     arg1, #16
                        abs     arg1, arg1              wc       ' get sign, force positive
                        muxc    t1, #1                           ' save sign

                        max     arg1, POSX16                     ' limit to 16-bit pos max
                        mov     t2, #0                           ' clear result workspace

:loop                   shr     arg2, #1                wc       ' get multiplier bit
        if_c            add     t2, arg1                         ' if 1, add muliplicand
                        shl     arg1, #1                wc       ' adjust for next bit
                        tjnz    arg2, #:loop                     ' again if more bits

                        test    t1, #1                  wc       ' recall and restore sign
                        negc    arg1, t2
                        shr     arg1, #16                        ' back to 16 bits
adjust_ret              ret

' -------------------------------------------------------------------------------------------------

ZERO                    long    0
MAX_VOL                 long    $1_0000
POSX16                  long    $0_7FFF

hx0000_FFFF             long    $0000_FFFF                       ' for truncating to 16 bits
hx8000_0000             long    $8000_0000                       ' zero level out at 32 bits

sdpntr                  long    0-0                              ' samples address for Stereo_Duty

pfpntr                  long    0-0                              ' pointer to playflag
srpntr                  long    0-0                              ' pointer to sample rate ticks
b0cntpntr               long    0-0                              ' pointer to buf0 count
b0pntr                  long    0-0                              ' pointer to buf0
b1cntpntr               long    0-0                              ' pointer to buf1 count
b1pntr                  long    0-0                              ' pointer to buf1
lvolpntr                long    0-0                              ' pointer to left volume setting
rvolpntr                long    0-0                              ' poitner to right volume setting

mode                    res     1                                ' mono or stereo

ratetix                 res     1                                ' ticks in sample rate
smpltimer               res     1                                ' timer for samples
bufid                   res     1                                ' buffer in use
bcpntr                  res     1                                ' buffer count pointer
bcount                  res     1                                ' samples in buffer
bufpntr                 res     1                                ' pointer to buffer in use
lchan                   res     1                                ' left ch sample
rchan                   res     1                                ' right ch sample
lvol                    res     1                                ' left volume
rvol                    res     1                                ' right volume

msamples                res     1

arg1                    res     1                                ' arguements for multiply
arg2                    res     1

t1                      res     1                                ' work vars
t2                      res     1

                        fit     496


dat { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}